#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#

import fixtures
import nanaimo
import nanaimo.version


def test_arg_version() -> None:
    """
    Verify --version argument contract.
    """
    expected = '.'.join(map(str, nanaimo.version.__version__))
    assert fixtures.run_nait(['--version']).stdout.decode('utf-8').startswith(expected)



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#
import argparse
import asyncio
import logging
import sys
import typing

import nanaimo


class _ArgparseSubparserArguments(nanaimo.Arguments):

    @classmethod
    def visit_argparse(cls,
                       test_class: typing.Type['nanaimo.NanaimoTest'],
                       subparsers: argparse._SubParsersAction,
                       loop: typing.Optional[asyncio.AbstractEventLoop] = None) -> None:
        subparser = subparsers.add_parser(test_class.__name__)  # type: 'argparse.ArgumentParser'
        subparser.add_argument('--test-timeout-seconds',
                               default='30',
                               type=float,
                               help='''Test will be killed and marked as a failure after
waiting for a result for this amount of time.''')
        test_class.on_visit_test_arguments(cls(subparser))
        subparser.set_defaults(func=test_class(loop))

    def __init__(self, subparser: argparse.ArgumentParser):
        self._subparser = subparser

    def add_argument(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        self._subparser.add_argument(*args, **kwargs)

    def set_defaults(self, **kwargs: typing.Any) -> None:
        self._subparser.set_defaults(**kwargs)


def _make_parser(loop: typing.Optional[asyncio.AbstractEventLoop] = None) -> argparse.ArgumentParser:
    """
        Defines the command-line interface. Provided as a separate factory method to
        support sphinx-argparse documentation.
    """

    epilog = '''**Example Usage**::

    python -m nanaimo -vv

----
'''

    parser = argparse.ArgumentParser(
        description='Run tests against hardware.',
        epilog=epilog,
        formatter_class=argparse.RawTextHelpFormatter)

    from nanaimo.version import __version__

    parser.add_argument('--version', action='version', version='.'.join(map(str, __version__)))

    parser.add_argument('--verbose', '-v', action='count',
                        help='verbosity level (-v, -vv)')

    subparsers = parser.add_subparsers(help='sub-command help')

    import nanaimo.builtin  # noqa: F401

    for test in nanaimo.NanaimoTest.__subclasses__():
        # https://github.com/python/mypy/issues/5374
        _ArgparseSubparserArguments.visit_argparse(test, subparsers, loop)  # type: ignore

    return parser


def _setup_logging(args: argparse.Namespace) -> None:
    fmt = '%(name)s : %(message)s'
    level = {0: logging.WARNING, 1: logging.INFO,
             2: logging.DEBUG}.get(args.verbose or 0, logging.DEBUG)
    logging.basicConfig(stream=sys.stderr, level=level, format=fmt)
    if args.verbose is not None and args.verbose >= 3:
        logging.getLogger('asyncio').setLevel(logging.DEBUG)


def main() -> int:
    """
    CLI entrypoint for running Nanaimo tests.
    """

    loop = asyncio.get_event_loop()

    parser = _make_parser(loop)
    args = parser.parse_args()

    _setup_logging(args)

    if hasattr(args, 'func'):
        result = loop.run_until_complete(args.func(nanaimo.Namespace(args)))
        try:
            return int(result)
        except ValueError:
            print('Nanaimo tests must return an int result!')
            raise
    else:
        parser.print_usage()
        return -1



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#
import nanaimo
import typing
import pytest
import _pytest


class _PyTestArguments(nanaimo.Arguments):

    def __init__(self, testparser: _pytest.config.argparsing.Parser):
        self._testparser = testparser

    def add_argument(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        self._testparser.addoption(*args, **kwargs)

    def set_defaults(self, **kwargs: typing.Any) -> None:
        raise NotImplementedError('pytest plugin does not support setting defaults.')


def pytest_addoption(parser: _pytest.config.argparsing.Parser) -> None:
    # import nanaimo.builtin  # noqa: F401

    # args = _PyTestArguments(parser)
    # for test in nanaimo.NanaimoTest.__subclasses__():
    #     # https://github.com/python/mypy/issues/5374
    #     group = parser.getgroup(type(test).__name__)
    #     test.on_visit_test_arguments(args)
    pass


@pytest.fixture
def hardware(request: _pytest.fixtures.FixtureRequest) -> typing.Callable[[], str]:
    # TODO: this is just a placeholder
    def _hello() -> str:
        return "Hello Pytest"

    return _hello



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#

import fixtures
import fixtures.simulators
import pytest
import asyncio
import nanaimo.serial
import nanaimo.bkprecision


@pytest.fixture
def paths_for_test():  # type: ignore
    return fixtures.Paths(__file__)


@pytest.mark.asyncio
async def test_turn_off(paths_for_test: fixtures.Paths, event_loop: asyncio.AbstractEventLoop) -> None:
    fake_serial = fixtures.simulators.Serial(['OK'], '\r')
    with nanaimo.serial.ConcurrentUart(fake_serial, eol='\r') as serial_sim:
        bk = nanaimo.bkprecision.Series1900BUart(serial_sim)
        await bk.turn_off()


@pytest.mark.asyncio
async def test_turn_on(paths_for_test: fixtures.Paths, event_loop: asyncio.AbstractEventLoop) -> None:
    fake_serial = fixtures.simulators.Serial(['OK'], '\r')
    with nanaimo.serial.ConcurrentUart(fake_serial, eol='\r') as serial_sim:
        bk = nanaimo.bkprecision.Series1900BUart(serial_sim)
        await bk.turn_on()


@pytest.mark.asyncio
@pytest.mark.timeout(10)
async def test_turn_on_timeout(paths_for_test: fixtures.Paths, event_loop: asyncio.AbstractEventLoop) -> None:
    fake_serial = fixtures.simulators.Serial(['NOPE'], '\r', loop_fake_data=False)
    with nanaimo.serial.ConcurrentUart(fake_serial, eol='\r') as serial_sim:
        bk = nanaimo.bkprecision.Series1900BUart(serial_sim, 1)
        with pytest.raises(asyncio.TimeoutError):
            await bk.turn_on()


@pytest.mark.asyncio
async def test_get_display(paths_for_test: fixtures.Paths, event_loop: asyncio.AbstractEventLoop) -> None:
    fake_serial = fixtures.simulators.Serial(['030201451', 'OK'], '\r', loop_fake_data=True)
    with nanaimo.serial.ConcurrentUart(fake_serial, eol='\r') as serial_sim:
        bk = nanaimo.bkprecision.Series1900BUart(serial_sim)
        display = await bk.get_display()
        assert 3.02 == display[0]
        assert 1.45 == display[1]
        assert nanaimo.bkprecision.Series1900BUart.StatusCC == display[2]



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#
import asyncio
import contextlib
import logging
import pathlib
import re
import typing

import nanaimo.serial


class Series1900BUart:
    """
    Control of a 1900B series BK Precision power supply via UART.
    """

    DefaultCommandTimeoutSeconds = 6.0

    # cspell: disable
    CommandTurnOn = 'SOUT0'
    CommandTurnOff = 'SOUT1'
    CommandGetDisplay = 'GETD'

    ResultOk = 'OK'

    StatusCV = 0
    StatusCC = 1

    CommandHelp = {
        'SOUT0': 'Switch on supply output.',
        'SOUT1': 'Switch off supply output.',
        'GETD': 'Read parameters displayed on the front panel of the supply.'
    }

    # cspell: enable

    @classmethod
    @contextlib.contextmanager
    def default_configuration(cls, port: typing.Union[str, pathlib.Path]) -> typing.Generator['Series1900BUart', None, None]:
        with nanaimo.serial.ConcurrentUart.new_default(str(port), 9600) as bkserial:
            bk = cls(bkserial, cls.DefaultCommandTimeoutSeconds)
            yield bk

    def __init__(self,
                 uart: nanaimo.serial.ConcurrentUart,
                 command_timeout_seconds: float = 0,
                 debug: bool = False):
        self._uart = uart
        self._uart.eol = '\r'
        self._command_timeout_seconds = command_timeout_seconds
        self._uart.timeout_seconds = command_timeout_seconds
        self._logger = logging.getLogger(type(self).__name__)
        self._debug = debug

    async def send_command(self, command: str) -> typing.Optional[typing.Any]:
        if command == '1':
            return await self._do_command(self.CommandTurnOn)
        elif command == '0':
            return await self._do_command(self.CommandTurnOff)
        elif command == 'r':
            return await self._do_command('\r\r\r\r')
        elif command == '?':
            return await self._do_command(self.CommandGetDisplay)
        else:
            raise ValueError('command {} is not a valid Series1900BUart command.'.format(command))

    async def turn_on(self) -> None:
        await self._do_command(self.CommandTurnOn)

    async def turn_off(self) -> None:
        await self._do_command(self.CommandTurnOff)

    async def get_display(self) -> typing.Tuple[float, float, int]:
        display = await self._do_command(self.CommandGetDisplay)
        if display is None or len(display) < 8:
            raise RuntimeError('Failed to obtain a voltage.')
        voltage = int(display[0:4]) / 100.0
        current = int(display[4:8]) / 100.0
        status = int(display[8])
        return (voltage, current, status)

    async def _do_command(self, command: str) -> typing.Optional[typing.Any]:
        try:
            command_help = self.CommandHelp[command]
            is_command = True
            self._logger.debug('Sending command %s (help=%s)', command, command_help)
        except KeyError:
            self._logger.debug('Sending characters %s', re.sub('\\r', '<cr>', command))
            is_command = False

        puttime_secs = await self._uart.put_line(command + self._uart.eol)
        previous_line = None
        if is_command:
            start_time = self._uart.time()
            while True:
                now = self._uart.time()
                if self._command_timeout_seconds > 0 and now - start_time > self._command_timeout_seconds:
                    raise asyncio.TimeoutError()
                if self._debug:
                    self._logger.debug('Waiting for response to command %s put before time %f seconds', command, puttime_secs)
                received_line = await self._uart.get_line((self._command_timeout_seconds - (now - start_time) if self._command_timeout_seconds > 0 else 0))
                if self._debug:
                    self._logger.debug('At %f Got line: %s', received_line.timestamp_seconds, received_line)
                # The result has to be from after the put or
                # it's an old result from a buffer.
                if received_line.timestamp_seconds > puttime_secs and received_line == self.ResultOk:
                    break
                # Skip any empty lines the device sends back.
                if len(received_line) > 0:
                    previous_line = received_line
        return previous_line

    async def wait_for_voltage(self, is_min: bool, threshold_v: float) -> None:
        while True:
            display_tuple = await self.get_display()
            voltage = display_tuple[0]
            if (voltage >= threshold_v if is_min else voltage <= threshold_v):
                if is_min:
                    self._logger.debug('---------------POWER SUPPLY UP----------------')
                else:
                    self._logger.debug('--------------POWER SUPPLY DOWN---------------')
                break
            await asyncio.sleep(.1)



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#
"""
Built-in :class:`NanaimoTest` for common scenarios.
"""
import asyncio
import pathlib

import nanaimo
import nanaimo.serial
import nanaimo.jlink


class GTestOverJLink(nanaimo.NanaimoTest):

    @classmethod
    def on_visit_test_arguments(cls, arguments: nanaimo.Arguments) -> None:
        nanaimo.serial.AbstractSerial.on_visit_test_arguments(arguments)
        nanaimo.jlink.ProgramUploaderJLink.on_visit_test_arguments(arguments)

    async def __call__(self, args: nanaimo.Namespace) -> int:
        import nanaimo
        import nanaimo.serial
        import nanaimo.gtest
        import nanaimo.jlink

        uploader = nanaimo.jlink.ProgramUploaderJLink()
        jlink_scripts = pathlib.Path(args.base_path).glob(args.jlink_scripts)
        parser = nanaimo.gtest.Parser(args.test_timeout_seconds)

        result = 0
        with nanaimo.serial.ConcurrentUart.new_default(args.port, args.port_speed) as monitor:
            for script in jlink_scripts:
                if result != 0:
                    break
                result = await asyncio.wait_for(uploader.upload(script), timeout=args.upload_timeout_seconds)
                if result != 0:
                    break
                result = await parser.read_test(monitor)

        return result



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#

import abc
import logging
import asyncio
import typing
import math


class AssertionError(RuntimeError):
    """
    Thrown by Nanaimo tests when an assertion has failed.
    """
    pass


class Arguments(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def add_argument(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        ...

    @abc.abstractmethod
    def set_defaults(self, **kwargs: typing.Any) -> None:
        ...


class Namespace:

    def __init__(self, parent: typing.Any):
        for name in vars(parent):
            setattr(self, name, getattr(parent, name))

    def __getattr__(self, key: str) -> typing.Any:
        return self.__dict__[key]

    def __eq__(self, other: typing.Any) -> bool:
        return vars(self) == vars(other)

    def __contains__(self, key: str) -> typing.Any:
        return key in self.__dict__


class NanaimoTest(metaclass=abc.ABCMeta):

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        return self._loop

    @classmethod
    @abc.abstractmethod
    def on_visit_test_arguments(cls, arguments: Arguments) -> None:
        ...

    @abc.abstractmethod
    async def __call__(self, args: Namespace) -> int:
        ...

    def __init__(self, loop: typing.Optional[asyncio.AbstractEventLoop] = None):
        self._logger = logging.getLogger(type(self).__name__)
        self._loop = (loop if loop is not None else asyncio.get_event_loop())

    async def countdown_sleep(self, sleep_time_seconds: float) -> None:
        """
        Calls `asyncio.sleep` once per-second for :param:`sleep_time_seconds` info
        logging a count-down. This can be used for long waits as an indication that
        the process is not deadlocked.
        """
        count_down = sleep_time_seconds
        while count_down >= 0:
            self._logger.info('%d', math.ceil(count_down))
            await asyncio.sleep(1)
            count_down -= 1

    async def _observe_tasks(self,
                             observer_co_or_f: typing.Union[typing.Coroutine, asyncio.Future],
                             timeout_seconds: float,
                             *args: typing.Union[typing.Coroutine, asyncio.Future]) -> \
            typing.Tuple[typing.Set[asyncio.Future], typing.Set[asyncio.Future]]:

        observing_my_future = asyncio.ensure_future(observer_co_or_f)
        the_children_are_our_futures = [observing_my_future]
        for co_or_f in args:
            the_children_are_our_futures.append(asyncio.ensure_future(co_or_f))

        start_time = self._loop.time()
        wait_timeout = (timeout_seconds if timeout_seconds > 0 else None)

        while True:
            done, pending = await asyncio.wait(
                the_children_are_our_futures,
                timeout=wait_timeout,
                return_when=asyncio.FIRST_COMPLETED)

            if observing_my_future.done():
                return done, pending

            if wait_timeout is not None and self._loop.time() - start_time > wait_timeout:
                break

        raise asyncio.TimeoutError()

    async def observe_tasks_assert_not_done(self,
                                            observer_co_or_f: typing.Union[typing.Coroutine, asyncio.Future],
                                            timeout_seconds: float,
                                            *args: typing.Union[typing.Coroutine, asyncio.Future]) -> typing.Set[asyncio.Future]:
        """
        Allows running a set of tasks but returning when an observer task completes. This allows a pattern where
        a single task is evaluating the side-effects of other tasks as a gate to continuing the test.
        """
        done, pending = await self._observe_tasks(observer_co_or_f, timeout_seconds, *args)
        if len(done) > 1:
            raise AssertionError('Tasks under observation completed before the observation was complete.')
        return pending

    async def observe_tasks(self,
                            observer_co_or_f: typing.Union[typing.Coroutine, asyncio.Future],
                            timeout_seconds: float,
                            *args: typing.Union[typing.Coroutine, asyncio.Future]) -> typing.Set[asyncio.Future]:
        """
        Allows running a set of tasks but returning when an observer task completes. This allows a pattern where
        a single task is evaluating the side-effects of other tasks as a gate to continuing the test.
        """

        done, pending = await self._observe_tasks(observer_co_or_f, timeout_seconds, *args)
        return pending



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#

import asyncio
import os

import pytest

import fixtures
import fixtures.simulators
import nanaimo
import nanaimo.gtest
import nanaimo.jlink
import nanaimo.serial


@pytest.mark.timeout(10)
def test_uart_monitor() -> None:
    """
    Verify the nanaimo.ConcurrentUart class using a mock serial port.
    """
    serial = fixtures.simulators.Serial(fixtures.FAKE_TEST_SUCCESS)
    last_line = fixtures.FAKE_TEST_SUCCESS[-1]
    with nanaimo.serial.ConcurrentUart(serial) as monitor:
        while True:
            line = monitor.readline()
            if line is None:
                os.sched_yield()
                continue
            elif line == last_line:
                break


@pytest.mark.asyncio
async def test_program_uploader() -> None:
    uploader = nanaimo.jlink.ProgramUploaderJLink(fixtures.get_mock_JLinkExe())
    assert 0 == await uploader.upload(fixtures.get_s32K144_jlink_script())


@pytest.mark.asyncio
async def test_program_uploader_failure() -> None:
    uploader = nanaimo.jlink.ProgramUploaderJLink(fixtures.get_mock_JLinkExe(), ['--simulate-error'])
    assert 0 != await uploader.upload(fixtures.get_s32K144_jlink_script())


@pytest.mark.asyncio
async def test_program_while_monitoring() -> None:
    scripts = fixtures.get_s32K144_jlink_scripts()
    uploader = nanaimo.jlink.ProgramUploaderJLink(fixtures.get_mock_JLinkExe())
    serial = fixtures.simulators.Serial(fixtures.FAKE_TEST_SUCCESS)
    uploads = 0
    with nanaimo.serial.ConcurrentUart(serial) as monitor:
        for script in scripts:
            serial.reset_fake_input()
            results = await asyncio.gather(
                nanaimo.gtest.Parser(10).read_test(monitor),
                uploader.upload(script)
            )
            assert 2 == len(results)

            for result in results:
                assert 0 == result
            uploads += 1
    assert uploads > 1


@pytest.mark.asyncio
async def test_failed_test() -> None:
    serial = fixtures.simulators.Serial(fixtures.FAKE_TEST_FAILURE)
    with nanaimo.serial.ConcurrentUart(serial) as monitor:
        assert 1 == await nanaimo.gtest.Parser(10).read_test(monitor)


@pytest.mark.asyncio
async def test_timeout_while_monitoring() -> None:
    serial = fixtures.simulators.Serial(['gibberish'], loop_fake_data=False)
    with nanaimo.serial.ConcurrentUart(serial) as monitor:
        assert 0 != await nanaimo.gtest.Parser(4.0).read_test(monitor)


@pytest.mark.timeout(10)
@pytest.mark.asyncio
async def test_observe_tasks(event_loop: asyncio.AbstractEventLoop) -> None:
    """
    Test the observe_tasks method of NanaimoTest
    """

    subject = fixtures.DummyNanaimoTest(event_loop)

    async def evaluating() -> int:
        return 0

    async def running() -> int:
        waits = 2
        while waits > 0:
            await asyncio.sleep(.1)
            waits -= 1
        return 1

    result = await subject.observe_tasks_assert_not_done(evaluating(),
                                                         0,
                                                         running())
    assert len(result) == 1
    should_be_running = result.pop()

    assert not should_be_running.done()

    assert 1 == await should_be_running


@pytest.mark.timeout(10)
@pytest.mark.asyncio
async def test_observe_tasks_failure(event_loop: asyncio.AbstractEventLoop) -> None:
    """
    Test the observe_tasks method of NanaimoTest where the running tasks exit.
    """

    subject = fixtures.DummyNanaimoTest(event_loop)

    async def evaluating() -> int:
        waits = 2
        while waits > 0:
            await asyncio.sleep(.1)
            waits -= 1
        return 1

    async def running() -> int:
        return 1

    with pytest.raises(nanaimo.AssertionError):
        await subject.observe_tasks_assert_not_done(evaluating(),
                                                    0,
                                                    running())


@pytest.mark.timeout(10)
@pytest.mark.asyncio
async def test_observe_tasks_failure_no_assert(event_loop: asyncio.AbstractEventLoop) -> None:
    """
    Test the observe_tasks method of NanaimoTest where the running tasks exit but without throwing
    an assertion error.
    """

    subject = fixtures.DummyNanaimoTest(event_loop)

    async def evaluating() -> int:
        waits = 2
        while waits > 0:
            await asyncio.sleep(.1)
            waits -= 1
        return 1

    async def running() -> int:
        return 1

    result = await subject.observe_tasks(evaluating(),
                                         0,
                                         running())

    assert 0 == len(result)


@pytest.mark.timeout(10)
@pytest.mark.asyncio
async def test_observe_tasks_timeout(event_loop: asyncio.AbstractEventLoop) -> None:
    """
    Test the observe_tasks method of NanaimoTest where the running tasks do not exit.
    """

    subject = fixtures.DummyNanaimoTest(event_loop)

    async def evaluating() -> int:
        while True:
            await asyncio.sleep(1)

    async def running() -> int:
        while True:
            await asyncio.sleep(1)

    with pytest.raises(asyncio.TimeoutError):
        await subject.observe_tasks_assert_not_done(evaluating(),
                                                    1,
                                                    running())


@pytest.mark.timeout(20)
@pytest.mark.asyncio
async def test_countdown_sleep(event_loop: asyncio.AbstractEventLoop) -> None:
    """
    Test the observe_tasks method of NanaimoTest where the running tasks do not exit.
    """
    subject = fixtures.DummyNanaimoTest(event_loop)

    await subject.countdown_sleep(5.3)



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#


def test_test_hardware_fixture(testdir) -> None:  # type: ignore
    testdir.makeconftest(
        """
# pytest_plugins = "nanaimo.pytest_plugin"
""")
    testdir.makepyfile("""
def test_hardware_fixture(hardware):
    assert hardware() == "Hello Pytest"
""")

    # run all tests with pytest
    result = testdir.runpytest()

    # check that all 4 tests passed
    result.assert_outcomes(passed=1)
